import pandas as pd


def get_value_counts_with_expansion_factor(
    df_, counted_cols, factor_column, normalize=True, dropna=True):
    '''
    Analogue to Series.value_counts but aplifiying rows by an expansion 
    factor column.
    
    Parameters
    ----------
    df : DataFrame,
        where expanded value counts will be performed.
    counted_cols : str or list of str
        column name(s) of column(s) to be counted.
    factor_column : str,
        column name specifying where the expansion factor of each row 
        should be found.
    '''
    
    if not isinstance(counted_cols, list):
        
        counted_cols = [counted_cols] 
                
        
    df = df_[counted_cols + [factor_column]].copy()
    
    if (not dropna) and (df[counted_cols].select_dtypes('category').shape[1] == 0):
        df[counted_cols] = df[counted_cols].fillna('NA')

    
    expanded_value_counts = df.groupby(counted_cols)[factor_column].sum()
    
    if normalize:
        expanded_value_counts = \
            expanded_value_counts / expanded_value_counts.sum()
        
    return expanded_value_counts

def get_percentage_table_with_expansion_factor(df, counted_col, factor_column,
                                               dropna=True):
    '''
    Analogue to Series.value_counts but aplifiying rows by an expansion 
    factor column, but additionally concatenates percentage column(s).
    
    Parameters
    ----------
    df : DataFrame,
        where expanded value counts will be performed.
    counted_cols : str or list of str
        column name(s) of column(s) to be counted.
    factor_column : str,
        column name specifying where the expansion factor of each row 
        should be found.
    '''
    
    sum_srs = get_value_counts_with_expansion_factor(
        df, counted_col, factor_column, normalize=False, dropna=dropna
        ).sort_index()
    sum_srs.name = 'exp_count'
    
    perc_srs = 100 * sum_srs / sum_srs.sum()
    perc_srs.name = 'percentage'
    perc_df = pd.concat([sum_srs, perc_srs], axis=1)

    # get percentage for the union of all levels except the last one 
    # which is already calculated 
    if isinstance(counted_col, list):
        
        union_str = '_U_'
        level_str = 'L'
        
        for l in range(len(counted_col) - 1, 0, -1):
            
            # initialize column which will contain Union(range(i)) 
            # percentage
            colname_ = union_str.join(
                [f"{level_str}{i}" for i in range(l)]) + '_percentage'
            perc_df.loc[:, colname_] = perc_df.percentage
            
            # get percentage subtotal on levels range(l)
            subtotals_srs = perc_df.groupby(level=list(range(l)))[
                'percentage'].sum()
            
            for idx, val in subtotals_srs.iteritems():
                
                # calculate Union(range(i)) percentage
                perc_df.loc[idx, colname_] = (100 * perc_df['percentage'].loc[
                    idx] / val).values
        
        # rename percentage column
        colname_ = union_str.join(
            [f"{level_str}{i}" for i in range(len(counted_col))]) +\
            '_percentage'
        perc_df = perc_df.rename(columns={'percentage': colname_})

    
    return perc_df


def get_weighted_average(df, col_to_average, weight_col):
    '''
    Computes the weighted average of pandas column using weitghts from
    another column

    Parameters
    ----------
    df : DataFrame,
        where columns are stored.
    col_to_average : hashable,
        points to the column from which the average is computed
    weight_col : hashable,
        point to the column where weights are extracted
    '''
    
    numerator = df[col_to_average] * df[weight_col]
    denominator = df[weight_col].sum()
    
    unpacked_division = numerator / denominator
    
    return unpacked_division.sum()



import pandas as pd
import numpy as np


def copy_docstring(sender):
    """Decorator for copying docstrings among functions.

    Parameters
    ----------
    sender: callable,
        where docstring will be copied from.
    receiver: callable,
        where dostring will be copied to.

    Example
    -------
    It can be used for different loader functions defined in each
    country source folder, normally in `utils.dataload`, by building
    wrappers of the agnostic functions defined in `covidsource.utils.
    dataload`.
    In the following example `sender=load_survey_data` and `receiver=
    load_survey_data_sv`:
    >>>@copy_docstring(load_survey_data)
    ...def load_survey_data_sv(path=SURVEY_DATA_PATH,
    ...                        load_func=LOCAL_LOAD_FUNC,
    ...                        columnnames_to_lower=True,
    ...                       ) -> pd.DataFrame:
    ...    return load_survey_data(
    ...        path, load_func,
    ...        columnnames_to_lower=columnnames_to_lower)

    source: https://softwareengineering.stackexchange.com/questions/
    386755/sharing-docstrings-between-similar-functions
    """

    def wrapper(receiver):
        receiver.__doc__ = sender.__doc__
        return receiver
    return wrapper


def bernoulli_on_rows(srs, row_bools, prob, seed = 300):
    '''Set values to zero using Bernoulli distribution on specific rows 
    of a Series.
    
    Parameters
    ----------
    srs : Series, 
        Series where specific rows are set as zero with probability `prob`.
    rows_bools: Series of bool
        specifies the columns where the values should be set to zero with
        probability `prob`.
    prob : float, 
        governs Bernoulli distribution.
    seed : random seed,
        used as argument of `np.random.seed` for ensuring replicability`.
    '''
    
    # generate bernoulli sample 
    np.random.seed(seed)
    bernoulli_srs = pd.Series(np.random.binomial(n = 1, p = prob, size =len(srs)))
    
    # exclude rows
    bernoulli_srs.loc[~ row_bools] = 1
    output = srs * bernoulli_srs.values
    
    return output
    

def uniform_decay(df, col, row_bools, decay):
    '''Apply uniform decay on column for specific rows. 
    
    Parameters
    ----------
    col : str,
        defines which is the column where this decay is applied.
    row_bools : Series of bool,
        define wich are the rows of `col` where uniform decay is  
        applied.
    decay : float,
        selected rows from `col` are multiplied by (1 - `decay`).
    '''
    unif_decay = row_bools.astype(float)
    unif_decay = unif_decay.replace(1, 1 - decay).replace(0, 1)
    output = df[col] * unif_decay
    return output
    

def proportional_cut(X, continuous_feature, threshold_col, n_buckets_threshold,
                     n_buckets_total, collapse_negative=False, **kwargs):
    """
    Transform continuous feature in binarized buckets proportional to a
    selected threshold.

    Parameters
    ----------
    X : pd.DataFrame,
        where columns are extracted.
    continuous_feature : column label,
        indicates wich columns is to be binarized.
    threshold_col : column label,
        indicates which column is used as threshold.
    n_buckets_threshold : int,
        number of bins to be used to reach threshold.
    n_buckets_total: int,
        total number of buckets. The rightmost bin will be unbounded,
        i.e. will contain which are greater than the seond last bin.
    collapse_negative : bool,
        whether negative values should be assigned to the leftmost bin.
        Default `False` (recommended).
    """

    bins = [i / n_buckets_threshold for i in range(0, n_buckets_total)]
    bins.append(np.inf)

    if collapse_negative:
        X.loc[X[continuous_feature] < 0, continuous_feature] = 0

    return pd.cut(
        X[continuous_feature] / X[threshold_col],
        bins=bins,
        labels=[i + 1 for i in range(n_buckets_total)],
        **kwargs
    )


def weighted_qcut(X, to_order_col, weights_col, q, **kwargs):
    """Return weighted quantile cuts from a given series, values.

    Parameters
    ----------
    X : pd.DataFrame,
        contains `to_order_col` and `weights_col`.
    to_order_col : column label,
        indicates the column according to which rows have to be ordered.
    weights_col : column label,
        indicates the column of weights.
    q : int,
        number of bins.
    """
    from pandas._libs.lib import is_integer

    if is_integer(q):
        quantiles = np.linspace(0, 1, q + 1)
    else:
        quantiles = q

    ordered_cum_sum = X[weights_col].iloc[X[to_order_col].argsort()].cumsum()
    bins = pd.cut(ordered_cum_sum / ordered_cum_sum.iloc[-1], quantiles,
                  **kwargs)
    return bins.sort_index()


def get_partition_bool_columns_dict(
        bool_cols, more_than_once, default, prefix='Solo ', suffix_to_remove='first'):
    """
    Returns dict to be used as `select_map` dict value on a
    `SelectTransformer` to partition a set of columns.

    The resulting partition first option will be two or more bool
    conditions from `bool_cols` followed by one per each bool column
    from `bool_cols` and if any bool condition is met from `bool_cols`,
    the default value will be used.

    Parameters
    ----------
    bool_cols : iterable of str,
        contains labels of columns where program `bool` observations
        will indicate whether or not the household is beneficiary of the
        program.
    more_than_once: str,
        used as choice when more than one bool condition is met on a
        row.
    default : str,
        used as default choice when no condition is met.
    prefix : str (optional),
        used as prefix on every choice which correspond to a single
        bool column from `bool_cols`. Default `'Solo '`
    suffix_to_remove : str (optional),
        to be removed from program cols on `choicelist`. Default
        `'first'`.
    """

    conditions = [
        lambda df, bool_cols=bool_cols:(df[bool_cols].sum(1) >= 2),
        * [(lambda df, col=col: df[col] == 1) for col in bool_cols
           ]]

    choices = [more_than_once, *[
        f'{prefix}{chn.replace(suffix_to_remove, "").replace("_", " ").strip()}'
        for chn in bool_cols
    ]]

    return {**dict(zip(choices, conditions)), **{'default': default}}


def new_poverty(original_poverty, modified_poverty):
    """
    Hardcoded function for getting `select_map` dict-value for computing
    new poverty.

    Parameters
    ----------
    original_poverty : column label,
        indicating the column of bools where reference poverty status
        needs to be taken.
    modified_poverty : column label,
        indicating column of bools where target poverty status  needs to
        be taken.
    """
    return {
        "Pobreza por COVID-19": lambda df, original_poverty=original_poverty,
        modified_poverty=modified_poverty: \
        (~ df[original_poverty]) & df[modified_poverty],
        "Se mantiene en pobreza": lambda df,
        original_poverty=original_poverty, modified_poverty=modified_poverty:
        df[original_poverty] & df[modified_poverty],
        "Se mantiene en no pobreza": lambda df,
        original_poverty=original_poverty, modified_poverty=modified_poverty:
        (~ df[original_poverty]) & (~ df[modified_poverty])
    }



import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from IPython.core.display import display


def unlimit_display_option(option, *args):
    """
    Wrapper for display function. It displays a DataFrame temporally
    setting an option to infinite.

    Parameters
    ----------
    option : str,
        indicates the option to be set to infinite.
    args : positional arguments,
        used as arguments of the `display` function.
    """
    current_option_value = pd.get_option(option)
    pd.set_option(option, None)
    display(*args)
    pd.set_option(option, current_option_value)


def cdisplay(*args):
    """
    Wrapper for display function. It displays  all columns of DataFrame
    objects present in `*args`
    """
    unlimit_display_option('display.max_columns', *args)


def rdisplay(*args):
    """
    Wrapper for display function. It displays  all columns of DataFrame
    objects present in `*args`
    """
    unlimit_display_option('display.max_rows', *args)

    
def percentage_count_plot(plot_srs, labels):
    """
    Draws barplot with percentages annotations calculated from
    `plot_srs`.

    Parameters
    ----------
    plot_srs : pd.Series,
        contains heighs of bars to be drawn. Percentages are calculated
        from it.
    labels : iterable
        contains the labels to be used as ticklabels on the x axis.
    """
    
    fig, ax = plt.subplots(figsize=(6, 4))
    
    ax.bar(x=labels, height=plot_srs, 
            color=sns.color_palette(n_colors=len(plot_srs)))

    plot_srs_pct = plot_srs / plot_srs.sum()

    # Add this loop to add the annotations
    for i, (p, l) in enumerate(zip(ax.patches, labels)):
        width, height = p.get_width(), p.get_height()
        x, y = p.get_xy() 
        ax.annotate('{:.1%}'.format(plot_srs_pct.iloc[i]), 
                    (x + width / 2, y + height + 0.01), 
                    ha='center', va='bottom')
        p.set_label(l)
    plt.legend(bbox_to_anchor=(1., 1))
    
    return fig, ax


import pandas as pd
import numpy as np
import warnings
from sklearn.base import BaseEstimator, TransformerMixin


class DropTransformer(BaseEstimator, TransformerMixin):
    """Class for the transformations of inputs into a useful representations
    for the Decision Support Platform.

    Attributes
    ----------
    drop_dict : dict of {str: list}
        contains column names as keys and a list of values that when observed the entire row is dropped.
        Example: dict = {'VD2002': [15, 16, 17]} drops rows which have VD2002 = 15, 16, or 17.
    """

    def __init__(self, drop_dict):
        self.col_drop_dict = drop_dict

    def fit(self, X, y = None):
        return self

    def transform(self, X, y = None):
        """
        """
        X_ = X.copy()

        for colname, drop_list in self.col_drop_dict.items():
            drop_rows = X_[colname].isin(drop_list)
            X_ = X_.drop(X_[drop_rows].index)

        return X_


class AggregateTransformerMixin:
    """
    Mixin class for aggregating rows of the DataFrame using `~.groupby`,
    according to `aggregate_map`

    Attributes
    ----------
    aggregate_map : dict of function labels -> iterable of column labels,
        all columns in the iterable are aggregated using its respective
        function. Function labels are added as suffixes for resulting
        columns to avoid name collisions.
    keep_first : iterable of column labels
        all columns referenced in this iterable are aggregated by using
        the first observation of each group
    groupby_: column label or iterable of column labels,
        used as arguments for `~.groupby`.
    """

    def __init__(self, aggregate_map, keep_first, groupby_):
        self.aggregate_map = aggregate_map
        # keep first is separated from aggregate_map because of
        # https://github.com/pandas-dev/pandas/issues/19598
        self.keep_first = keep_first
        self.groupby_ = groupby_

    def aggregate_transform(self, X):
        X_ = X.copy()

        groupped_X = X_.groupby(self.groupby_)
        agg_df_list = []

        # process keep first
        agg_df_list.append(groupped_X[self.keep_first].nth(0).add_suffix('_first'))

        for func_name, col_list in self.aggregate_map.items():
            # parse agggregate dict
            iter_agg_dict = {col: func_name for col in col_list}

            # calculate aggregated
            agg_df_list.append(groupped_X.agg(iter_agg_dict).add_suffix(f'_{func_name}'))

        ret = pd.concat(agg_df_list, axis=1).reset_index()

        return ret


class AggregateTransformer(AggregateTransformerMixin, BaseEstimator, TransformerMixin):
    """
    Class for aggregating rows of the DataFrame using `~.groupby`,
    according to `aggregate_map`.

    Attributes
    ----------
    aggregate_map : dict of function labels -> iterable of column labels,
        all columns in the iterable are aggregated using its respective
        function. Function labels are added as suffixes for resulting
        columns to avoid name collisions.
    keep_first : iterable of column labels
        all columns referenced in this iterable are aggregated by using
        the first observation of each group
    groupby_: column label or iterable of column labels,
        used as arguments for `~.groupby`.
    """

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return self.aggregate_transform(X)




import pandas as pd
import boto3


def load_survey_data(path, load_func,
                     columnnames_to_lower=True,
                     ) -> pd.DataFrame:
    """
    Loads survey DataFrame locally

    Parameters
    ----------
    path : str or pathlib.Path,
        dataframe filename.
    columnnames_to_lower : bool,
        specifies if column names need to be converted to lower.
    load_func : callable,
        it's unique argument needs to be `path`.
    """
    survey_hp = load_func(path)

    if columnnames_to_lower:
        survey_hp.columns = survey_hp.columns.str.lower()

    return survey_hp


def load_s3_data(profile_name,
                 bucket_name,
                 key,
                 load_func,
                 columnnames_to_lower=True):
    """
    Loads data DataFrame from s3 bucket

    Parameters
    ----------
    profile_name : str ,
        AWS profile name.
    bucket_name : str ,
        AWS bucket name.
    key : str,
        AWS key to be used on bucket to find the target file
    columnnames_to_lower : bool,
        specifies if column names need to be converted to lower.
    load_func : callable,
        it's unique argument needs to be `path`.
    """
    # instance s3 connection
    boto3.setup_default_session(profile_name=profile_name)
    s3 = boto3.client('s3')

    # read file
    obj = s3.get_object(Bucket=bucket_name, Key=key)
    ar_df = load_func(obj)

    if columnnames_to_lower:
        ar_df.columns = ar_df.columns.str.lower()

    return ar_df


import pandas as pd
import numpy as np
import warnings
from sklearn.base import BaseEstimator, TransformerMixin


class MapTransformer(BaseEstimator, TransformerMixin):
    """Class for the transformations of inputs into a useful representations 
    for the Decision Support Platform.

    Attributes
    ----------
    names_map : dict of {str: dict}
        contains column names as keys and values are dict's to be used
        as `pd.Series.map` argument.
    """

    def __init__(self, col_map_dict):
        self.col_map_dict = col_map_dict

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        """
        """
        # copy
        X_ = X.copy()

        for colname, map_dict in self.col_map_dict.items():

            X_[colname] = X_[colname].map(map_dict)

        return X_

class NameTransformer(BaseEstimator, TransformerMixin):
    """Class used to select and rename the columns of a DataFrame.

    Attributes
    ----------
    names_map : dict,
        defines how columns are renamed
    keep_features : list,
        defines columns which are kept with the original name
    """

    def __init__(self, names_map, keep_features):
        """
        Parameters
        ----------
        names_map : dict {str: str},
            defines how columns are renamed. Keys are original names and
            values, the new ones
        keep_features : list,
            defines columns which are kept with the original name.
        """

        intersection_ = set(names_map.keys()).intersection(keep_features)
        assert len(intersection_) == 0, \
            f"""The following columns are defined both in `keep_features` as in `names_map.keys()`:
            {intersection_}
            """

        self.names_map = names_map
        self.keep_features = keep_features


    def fit(self, X, y=None):

        return self


    def transform(self, X, y=None):
        """
        Transform dataset.
        Parameters
        ----------
        X : pandas DataFrame, shape=(n_samples, n_features)
            Input data to be transformed. Use ``dtype=np.float32`` for maximum
            efficiency.
        Returns
        -------
        X_transformed : pandas DataFrame, shape=(n_samples, n_out)
            Transformed dataset.
        """
        X_ = X[list(self.names_map.keys()) + self.keep_features].copy()
        X_ = X_.rename(self.names_map, axis=1)

        return X_




class SelectTransformerMixin:
    """Mixin class for applying `np.select` to DataFrames, using a dict as
    unique parameter

    Attributes
    ----------
    select_map: dict of {colname: dict of {`choicelist`: `condlist`}}
        where `choicelist` and `condlist` are the arguments of
        `np.select`. For setting `default` argument of `np.select`
         use the specific key `'default'` within the internal dict.
    """

    def __init__(self, select_map):

        self.select_map = select_map

    def select_transform(self, X):

        # copy
        X_ = X.copy()

        for col_name, conditions_dict in self.select_map.items():
            # debug = 'y_lab' in col_name

            # initialize np.select arguments
            default_value = conditions_dict.get('default')
            default_value = default_value(X_) if callable(default_value) else default_value
            cond_list = []
            choice_list = []

            #if debug:
                #affected = 0

            for k, func in conditions_dict.items():

                if k != 'default':

                    # collect conditions and choice labels
                    cond_list.append(func(X_))
                    # if debug:
                    #     affected += func(X_).sum()

                    if callable(k):
                        choice_list.append(k(X_))

                    else:
                        choice_list.append(k)
                    # print(k, func(X_).sum())
            # if debug:
            #     print(f'affected: {affected:,}')


            # insert new column
            X_.loc[:, col_name] = np.select(cond_list, choice_list, default=default_value)

        return X_


class SelectTransformer(SelectTransformerMixin, BaseEstimator, TransformerMixin):
    """
    Class for applying `np.select` to DataFrames, using a dict as
    unique parameter.
    Attributes
    ----------
    select_map: dict of {colname: dict of {`choicelist`: `condlist`}}
        where `choicelist` and `condlist` are the arguments of
        `np.select`. For setting `default` argument of `np.select`
         use the specific key `'default'` within the internal dict.
    """

    def fit(self, X, y=None):
        return self

    def transform(self, X):

        # apply transform method
        return self.select_transform(X)


class AssignTransformerMixin:
    """Mixin class for applying `DataFrame.assign`, using a dict as
    unique parameter

    Attributes
    ----------
    assign_map : dict of {colname: udf},
        `udf` must return a series which will be stored on the
        dataframe on `colname` column.

    NOTE: `udf` must never modify input DataFrame. Pandas does not
    check this!
    """

    def __init__(self, assign_map):

        self.assign_map = assign_map

    def assign_transform(self, X):

        # copy
        X_ = X.copy()

        return X_.assign(**self.assign_map)


class AssignTransformer(AssignTransformerMixin, BaseEstimator, TransformerMixin):
    """Class for applying `np.select` to DataFrames, using a dict as
    unique parameter.

    Attributes
    ----------
    assign_map : dict of {colname: udf},
        `udf` must return a series which will be stored on the
        dataframe on `colname` column.

    NOTE: `udf` must never modify input DataFrame. Pandas does not
    check this!
    """

    def fit(self, X, y=None):
        return self

    def transform(self, X):

        # apply transform method
        return self.assign_transform(X)


class AssSelTransformer(
    BaseEstimator, TransformerMixin, SelectTransformerMixin,
    AssignTransformerMixin):
    '''Generates new cols by user DataFrame.assign and by np.select

    Parameters
    ----------
    select_map: dict of {colname: dict of {`choicelist`: `condlist`}}
        where `choicelist` and `condlist` are the arguments of
        `np.select`. For setting `default` argument of `np.select`
         use the specific key `'default'` within the internal dict.

    assign_map : dict {colname: udf},
        `udf` must return a series which will be stored on the
        dataframe on `colname` column
    '''


    def __init__(self, select_map, assign_map):

        warnings.warn(
            "The `AssSelTransformer` class will be deprecated. Prefer with `AssignTransformer` and/or "
            "`SelectTransformer`",
            DeprecationWarning, stacklevel=2)

        SelectTransformerMixin.__init__(self, select_map)
        AssignTransformerMixin.__init__(self, assign_map)


    def transform(self, X):
        '''Calls all transform methods'''

        # copy
        X_ = X.copy()

        # apply inherited transform methods
        X_ = self.select_transform(X_)
        X_ = self.assign_transform(X_)

        return X_














import setuptools

with open("readme.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="criteria-etl",
    version="0.1.0",
    author="Prosperia Social",
    author_email="developers.etl@prosperia.ai",
    description="A library for Criteria related data wrangling.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/prosper-ia/covid-response-source",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=[
        "pandas>=1", "scikit-learn"
    ]
)


from pathlib import Path

# loading configurations
SURVEY_DATA_PATH = (Path(__file__).parent.parent.parent / None).resolve()
S3_PROFILE_NAME = None
S3_BUCKET_NAME = None
S3_KEY = None


def LOCAL_LOAD_FUNC(path): pass


def S3_LOAD_FUNC(obj): pass


# Global variables
key_variable_A = None
key_variable_B = None
